/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.apache.flink.streaming.tests;

import org.elasticsearch.action.update.UpdateRequest;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

/** Factory for creating UpdateRequests. */
public interface UpdateRequestFactory extends Serializable {
    UpdateRequest createUpdateRequest(KeyValue<Integer, String> element);

    /**
     * Utility to convert {@link KeyValue} elements into Elasticsearch-compatible format.
     *
     * @param element The element to be converted.
     * @return The map with the element's fields.
     */
    static Map<String, Object> prepareDoc(KeyValue<Integer, String> element) {
        Map<String, Object> json = new HashMap<>();
        json.put("key", element.key);
        json.put("value", element.value);
        return json;
    }
}
